#include "bat.hpp"
#include "environment.hpp"
#include "key_positions_structure.hpp"
#include "interpolation.hpp"
#include "terrain.hpp"

void bat::initialize_bat()
{
	
	// Initialize the temporary mesh_drawable that will be inserted in the hierarchy
	mesh_drawable body;
	mesh_drawable wing1;
	mesh_drawable wing2;

	// Create the geometry of the meshes
	//   Note: this geometry must be set in their local coordinates with respect to their position in the hierarchy (and with respect to their animation)

	body.initialize_data_on_gpu(mesh_load_file_obj(project::path + "assets/batbodyok.obj"));
	wing1.initialize_data_on_gpu(mesh_load_file_obj(project::path + "assets/batwingleft.obj"));
	wing2.initialize_data_on_gpu(mesh_load_file_obj(project::path + "assets/batwingright.obj"));


	// Scale the model
	//body.model.scaling = 0.2f;
	//wing1.model.scaling = 0.2f;
	//wing2.model.scaling = 0.2f;



	// Set the color of some elements
	body.material.color = { 0.3f,0.3f,0.3f };
	wing1.material.color = { 0.2f,0.2f,0.2f };
	wing2.material.color = { 0.2f,0.2f,0.2f };







	// Add the elements in the hierarchy
	//   The syntax is hierarchy.add(mesh_drawable, "name of the parent element", [optional: local translation in the hierarchy])
	//   Notes: 
	//     - An element must necessarily be added after its parent
	//     - The first element (without explicit name of its parent) is assumed to be the root.

	bat.add(body, "Bat base");
	bat.add(wing1, "Bat wing left1", "Bat base", { -0.05f,0,0 });
	bat.add(wing2, "Bat wing right1", "Bat base", { 0.05f,0,0 });


	N = 0;
	size.resize(N);
	isdead.resize(N);
	pos.resize(N);
	pos2.resize(N);
	keyframe.resize(N);
}



void bat::initialize_mvt(vec3 p, int m, float terrain_length)
{
	// Definition of the initial data
	//--------------------------------------//

	// Key 3D positions
	numarray<vec3> key_positions =
	{ {-2,1,0}, {0,1,0}, {2,1,0}, {1,3,0}, {2,3,0}, {2,3,1}, {2,2.5,1.5}, {1.5,0,1}, {1.5,0,-1}, {0,0,-1}, {0,-0.5,-1}, {-2,1,-1} };

	float h = 20 + evaluate_terrain_height(0, 0, terrain_length);
	for (int i = 0; i < key_positions.size(); i++) {
		key_positions[i] *= size[m];
		key_positions[i][2] += h;
		key_positions[i] += p;
	}

	// Key times (time at which the position must pass in the corresponding position)
	numarray<float> key_times =
	{ 0.0f, 1.0f, 2.0f, 2.5f, 3.0f, 3.5f, 3.75f, 4.5f, 5.0f, 6.0f, 7.0f, 8.0f };

	// Initialize the helping structure to display/interact with these positions
	keyframe[m].initialize(key_positions, key_times);


	// Set timer bounds
	// The timer must span a time interval on which the interpolation can be conducted
	// By default, set the minimal time to be key_times[1], and the maximal time to be key_time[N-2] (enables cubic interpolation)
	int NT = key_times.size();
	timer_mvt.t_min = key_times[0];
	timer_mvt.t_max = key_times[NT - 1];
	timer_mvt.t = timer_mvt.t_min;
}

void bat::display_mvt()
{

	// Update the current time
	timer_mvt.update();
	float t = timer_mvt.t;


	for (int m = 0; m < N; m++) {
		// clear trajectory when the timer restart
		if (t < timer_mvt.t_min + 0.1f)
			keyframe[m].trajectory.clear();

		if (t + 0.1f > timer_mvt.t_max) pos2[m] = interpolation(0.01f, keyframe[m].key_positions, keyframe[m].key_times, 0.5f);
		else pos2[m] = interpolation((t + 0.1f), keyframe[m].key_positions, keyframe[m].key_times, 0.5f);

		// Display the key positions and lines b/w positions
		//keyframe.display_key_positions(environment);

		// Compute the interpolated position
		//  This is this function that you need to complete
		pos[m] = interpolation(t, keyframe[m].key_positions, keyframe[m].key_times, 0.5f);

		// Display the interpolated position (and its trajectory)
		//keyframe.display_current_position(p, environment);
	}
}


void bat::add_bat(vec3 new_pos, float terrain_length) {
	N++;
	pos.resize(N);
	pos2.resize(N);
	isdead.resize(N);
	size.resize(N);
	keyframe.resize(N);


	pos[N - 1] = new_pos;
	isdead[N - 1] = false;
	size[N - 1] = rand_interval(0.05f, 1.0f);;
	initialize_mvt(new_pos, N - 1, terrain_length);

}

